"""FSW models for ground imaging scenarios."""

from typing import TYPE_CHECKING, Iterable

import numpy as np
from Basilisk.architecture import messaging
from Basilisk.fswAlgorithms import (
    locationPointing,
    scanningInstrumentController,
    simpleInstrumentController,
)

from bsk_rl.sim import dyn
from bsk_rl.sim.fsw import BasicFSWModel, SteeringFSWModel, Task, action
from bsk_rl.utils import vizard
from bsk_rl.utils.functional import default_args
from bsk_rl.utils.orbital import rv2HN

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.sim.dyn import DynamicsModel


class ImagingFSWModel(BasicFSWModel):
    """Extend FSW with instrument pointing and triggering control."""

    @classmethod
    def _requires_dyn(cls) -> list[type["DynamicsModel"]]:
        return super()._requires_dyn() + [dyn.ImagingDynModel]

    def __init__(self, *args, **kwargs) -> None:
        """Adds instrument pointing and triggering control to FSW."""
        super().__init__(*args, **kwargs)

    @property
    def c_hat_P(self):
        """Instrument pointing direction in the planet frame."""
        c_hat_B = self.locPoint.pHat_B
        return np.matmul(self.dynamics.BP.T, c_hat_B)

    @property
    def c_hat_H(self):
        """Instrument pointing direction in the hill frame."""
        c_hat_B = self.locPoint.pHat_B
        HN = rv2HN(self.satellite.dynamics.r_BN_N, self.satellite.dynamics.v_BN_N)
        return HN @ self.satellite.dynamics.BN.T @ c_hat_B

    def _make_task_list(self) -> list[Task]:
        return super()._make_task_list() + [self.LocPointTask(self)]

    def _set_gateway_msgs(self) -> None:
        super()._set_gateway_msgs()
        self.dynamics.instrument.nodeStatusInMsg.subscribeTo(
            self.insControl.deviceCmdOutMsg
        )

    class LocPointTask(Task):
        """Task to point at targets and trigger the instrument."""

        name = "locPointTask"

        def __init__(self, fsw, priority=96) -> None:  # noqa: D107
            """Task to point the instrument at ground targets."""
            super().__init__(fsw, priority)

        def _create_module_data(self) -> None:
            # Location pointing configuration
            self.locPoint = self.fsw.locPoint = locationPointing.locationPointing()
            self.locPoint.ModelTag = "locPoint"

            # SimpleInstrumentController configuration
            self.insControl = self.fsw.insControl = (
                simpleInstrumentController.simpleInstrumentController()
            )
            self.insControl.ModelTag = "instrumentController"

        def _setup_fsw_objects(self, **kwargs) -> None:
            self.setup_location_pointing(**kwargs)
            self.setup_instrument_controller(**kwargs)
            self.show_sensor()

        @default_args(inst_pHat_B=[0, 0, 1])
        def setup_location_pointing(
            self, inst_pHat_B: Iterable[float], **kwargs
        ) -> None:
            """Set the Earth location pointing guidance module.

            Args:
                inst_pHat_B: Instrument pointing direction.
                kwargs: Passed to other setup functions.
            """
            self.locPoint.pHat_B = inst_pHat_B
            self.locPoint.scAttInMsg.subscribeTo(
                self.fsw.dynamics.simpleNavObject.attOutMsg
            )
            self.locPoint.scTransInMsg.subscribeTo(
                self.fsw.dynamics.simpleNavObject.transOutMsg
            )
            self.locPoint.locationInMsg.subscribeTo(
                self.fsw.dynamics.imagingTarget.currentGroundStateOutMsg
            )
            self.locPoint.useBoresightRateDamping = 1
            messaging.AttGuidMsg_C_addAuthor(
                self.locPoint.attGuidOutMsg, self.fsw.attGuidMsg
            )

            self._add_model_to_task(self.locPoint, priority=1198)

        @default_args(imageAttErrorRequirement=0.01, imageRateErrorRequirement=None)
        def setup_instrument_controller(
            self,
            imageAttErrorRequirement: float,
            imageRateErrorRequirement: float,
            **kwargs,
        ) -> None:
            """Set the instrument controller parameters.

            The instrument controller is used to take an image when certain relative
            attitude requirements are met, along with the access requirements of the
            target (i.e. ``imageTargetMinimumElevation`` and ``imageTargetMaximumRange``
            as set in :class:`~bsk_rl.sim.dyn.ImagingDynModel.setup_imaging_target`).

            Args:
                imageAttErrorRequirement: [MRP norm] Pointing attitude error tolerance
                    for imaging.
                imageRateErrorRequirement: [rad/s] Rate tolerance for imaging. Disable
                    with ``None``.
                kwargs: Passed to other setup functions.
            """
            self.insControl.attErrTolerance = imageAttErrorRequirement
            if imageRateErrorRequirement is not None:
                self.insControl.useRateTolerance = 1
                self.insControl.rateErrTolerance = imageRateErrorRequirement
            self.insControl.attGuidInMsg.subscribeTo(self.fsw.attGuidMsg)
            self.insControl.locationAccessInMsg.subscribeTo(
                self.fsw.dynamics.imagingTarget.accessOutMsgs[-1]
            )

            self._add_model_to_task(self.insControl, priority=987)

        @vizard.visualize
        def show_sensor(self, vizInterface=None, vizSupport=None):
            """Visualize the sensor in Vizard."""
            genericSensor = vizInterface.GenericSensor()
            genericSensor.normalVector = self.locPoint.pHat_B
            genericSensor.r_SB_B = [0.0, 0.0, 0.0]
            genericSensor.fieldOfView.push_back(4 * self.insControl.attErrTolerance)
            genericSensor.color = vizInterface.IntVector(
                vizSupport.toRGBA255(self.fsw.satellite.vizard_color, alpha=0.5)
            )
            cmdInMsg = messaging.DeviceCmdMsgReader()
            cmdInMsg.subscribeTo(self.insControl.deviceCmdOutMsg)
            genericSensor.genericSensorCmdInMsg = cmdInMsg
            self.fsw.satellite.vizard_data["genericSensorList"] = [genericSensor]

        def reset_for_action(self) -> None:
            """Reset pointing controller."""
            self.fsw.dynamics.imagingTarget.Reset(self.fsw.simulator.sim_time_ns)
            self.locPoint.Reset(self.fsw.simulator.sim_time_ns)
            self.insControl.controllerStatus = 0
            return super().reset_for_action()

    @action
    def action_image(self, r_LP_P: Iterable[float], data_name: str) -> None:
        """Attempt to image a target at a location.

        This action sets the target attitude to one tracking a ground location. If the
        target is within the imaging constraints, an image will be taken and stored in
        the data buffer. The instrument power sink will be active as long as the task is
        enabled.

        Args:
            r_LP_P: [m] Planet-fixed planet relative target location.
            data_name: Data buffer to store image data to.
        """
        self.insControl.controllerStatus = 1
        self.dynamics.instrumentPowerSink.powerStatus = 1
        self.dynamics.imagingTarget.r_LP_P_Init = r_LP_P
        self.dynamics.instrument.nodeDataName = data_name
        self.insControl.imaged = 0
        self.simulator.enableTask(self.LocPointTask.name + self.satellite.name)

    @action
    def action_downlink(self) -> None:
        """Attempt to downlink data.

        This action points the satellite nadir and attempts to downlink data. If the
        satellite is in range of a ground station, data will be downlinked at the specified
        baud rate. The transmitter power sink will be active as long as the task is enabled.
        """
        self.hillPoint.Reset(self.simulator.sim_time_ns)
        self.trackingError.Reset(self.simulator.sim_time_ns)
        self.dynamics.transmitter.dataStatus = 1
        self.dynamics.transmitterPowerSink.powerStatus = 1
        self.simulator.enableTask(
            BasicFSWModel.NadirPointTask.name + self.satellite.name
        )
        self.simulator.enableTask(
            BasicFSWModel.TrackingErrorTask.name + self.satellite.name
        )


class ContinuousImagingFSWModel(ImagingFSWModel):
    """FSW model for continuous nadir scanning."""

    def __init__(self, *args, **kwargs) -> None:
        """FSW model for continuous nadir scanning.

        Instead of imaging point targets, this model is used to continuously scan the
        ground while pointing nadir.
        """
        super().__init__(*args, **kwargs)

    class LocPointTask(ImagingFSWModel.LocPointTask):
        """Task to point nadir and trigger the instrument."""

        def __init__(self, *args, **kwargs) -> None:
            """Task to point nadir and trigger the instrument."""
            super().__init__(*args, **kwargs)

        def _create_module_data(self) -> None:
            # Location pointing configuration
            self.locPoint = self.fsw.locPoint = locationPointing.locationPointing()
            self.locPoint.ModelTag = "locPoint"

            # scanningInstrumentController configuration
            self.insControl = self.fsw.insControl = (
                scanningInstrumentController.scanningInstrumentController()
            )
            self.insControl.ModelTag = "instrumentController"

        @default_args(imageAttErrorRequirement=0.01, imageRateErrorRequirement=None)
        def setup_instrument_controller(
            self,
            imageAttErrorRequirement: float,
            imageRateErrorRequirement: float,
            **kwargs,
        ) -> None:
            """Set the instrument controller parameters for scanning.

            As long as these two conditions are met, scanning will occur continuously.

            Args:
                imageAttErrorRequirement: [MRP norm] Pointing attitude error tolerance
                    for imaging.
                imageRateErrorRequirement: [rad/s] Rate tolerance for imaging. Disable
                    with None.
                kwargs: Passed to other setup functions.
            """
            self.insControl.attErrTolerance = imageAttErrorRequirement
            if imageRateErrorRequirement is not None:
                self.insControl.useRateTolerance = 1
                self.insControl.rateErrTolerance = imageRateErrorRequirement
            self.insControl.attGuidInMsg.subscribeTo(self.fsw.attGuidMsg)
            self.insControl.accessInMsg.subscribeTo(
                self.fsw.dynamics.imagingTarget.accessOutMsgs[-1]
            )

            self._add_model_to_task(self.insControl, priority=987)

        def reset_for_action(self) -> None:
            """Reset scanning controller."""
            self.instMsg = messaging.DeviceCmdMsg_C()
            self.instMsg.write(messaging.DeviceCmdMsgPayload())
            self.fsw.dynamics.instrument.nodeStatusInMsg.subscribeTo(self.instMsg)
            return super().reset_for_action()

    @action
    def action_nadir_scan(self) -> None:
        """Scan nadir.

        This action points the instrument nadir and continuously adds data to the buffer
        as long as attitude requirements are met. The instrument power sink is active
        as long as the action is set.
        """
        self.dynamics.instrument.nodeStatusInMsg.subscribeTo(
            self.insControl.deviceCmdOutMsg
        )
        self.insControl.controllerStatus = 1
        self.dynamics.instrumentPowerSink.powerStatus = 1
        self.dynamics.imagingTarget.r_LP_P_Init = np.array(
            [0, 0, 0.1]
        )  # All zero causes an error
        self.dynamics.instrument.nodeDataName = "nadir"
        self.simulator.enableTask(self.LocPointTask.name + self.satellite.name)

    @action
    def action_image(self, *args, **kwargs) -> None:
        """Disable ``action_image`` from parent class.

        :meta private:
        """
        raise NotImplementedError("Use action_nadir_scan instead")


class SteeringImagerFSWModel(SteeringFSWModel, ImagingFSWModel):
    """Convenience type for ImagingFSWModel with MRP steering."""

    def __init__(self, *args, **kwargs) -> None:
        """Convenience type that combines the imaging FSW model with MRP steering."""
        super().__init__(*args, **kwargs)


__doc_title__ = "Ground Imaging"
__all__ = [
    "ImagingFSWModel",
    "ContinuousImagingFSWModel",
    "SteeringImagerFSWModel",
]



from functools import partial

import gymnasium as gym
import numpy as np
import pytest
from pytest import approx

from bsk_rl import act, data, obs, sats
from bsk_rl.scene import UniformNadirScanning, UniformTargets
from bsk_rl.sim import dyn, fsw
from bsk_rl.utils.orbital import random_circular_orbit, random_orbit

#########################
# Composed Action Tests #
#########################


class TestImagingAndDownlink:
    class ImageSat(sats.ImagingSatellite):
        dyn_type = dyn.GroundStationDynModel
        fsw_type = fsw.ImagingFSWModel
        observation_spec = [obs.Time()]
        action_spec = [act.Downlink(), act.Image(n_ahead_image=10)]

    env = gym.make(
        "SatelliteTasking-v1",
        satellite=ImageSat(
            "EO-1",
            sat_args=ImageSat.default_sat_args(
                oe=random_orbit,
                imageAttErrorRequirement=0.05,
                imageRateErrorRequirement=0.05,
                instrumentBaudRate=1.0,
                dataStorageCapacity=3.0,
                transmitterBaudRate=-1.0,
            ),
        ),
        scenario=UniformTargets(n_targets=1000),
        rewarder=data.UniqueImageReward(),
        sim_rate=1.0,
        time_limit=10000.0,
        max_step_duration=1e9,
        disable_env_checker=True,
    )

    def test_image(self):
        self.env.reset()
        storage_init = self.env.unwrapped.satellite.dynamics.storage_level
        for i in range(10):
            self.env.step(i + 1)
        assert self.env.unwrapped.satellite.dynamics.storage_level > storage_init
        assert self.env.unwrapped.satellite.dynamics.storage_level == approx(3.0)

    @pytest.mark.flaky(retries=5, delay=1)
    def test_downlink(self):
        self.env.reset()
        for i in range(10):
            self.env.step(i + 1)
        storage_init = self.env.unwrapped.satellite.dynamics.storage_level
        assert storage_init > 0.0  # Should be filled from random stepping
        self.env.step(0)  # Should encounter a downlink opportunity before timeout
        assert self.env.unwrapped.satellite.dynamics.storage_level < storage_init

    def test_image_by_name(self):
        # Smoketest
        self.env.reset()
        target = self.env.unwrapped.satellite.find_next_opportunities(
            n=10, types="target"
        )[9]["object"]
        self.env.step(target)
        target = self.env.unwrapped.satellite.find_next_opportunities(
            n=10, types="target"
        )[9]["object"]
        self.env.step(target.id)
        assert True


###########################
# Individual Action Tests #
###########################


class TestChargingAction:
    class ChargeSat(sats.Satellite):
        dyn_type = dyn.BasicDynamicsModel
        fsw_type = fsw.BasicFSWModel
        observation_spec = [obs.Time()]
        action_spec = [act.Charge()]

    env = gym.make(
        "SatelliteTasking-v1",
        satellite=ChargeSat(
            "Charger",
            sat_args=ChargeSat.default_sat_args(
                # High, inclined orbit makes eclipse unlikely
                oe=random_orbit(a=50000, i=90),
                batteryStorageCapacity=500_000,
                storedCharge_Init=250_000,
            ),
        ),
        scenario=UniformTargets(n_targets=0),
        rewarder=data.NoReward(),
        sim_rate=1.0,
        max_step_duration=300.0,
        time_limit=300.0,
        disable_env_checker=True,
    )

    def test_charging_action(self):
        self.env.reset()
        init_charge = self.env.unwrapped.satellite.dynamics.battery_charge
        self.env.step(0)  # Charge
        assert self.env.unwrapped.satellite.dynamics.battery_charge > init_charge


class TestDesatAction:
    class DesatSat(sats.Satellite):
        dyn_type = dyn.BasicDynamicsModel
        fsw_type = fsw.BasicFSWModel
        observation_spec = [obs.Time()]
        action_spec = [act.Desat()]

    def make_env(self):
        return gym.make(
            "SatelliteTasking-v1",
            satellite=self.DesatSat(
                "Ellite",
                sat_args=self.DesatSat.default_sat_args(
                    oe=partial(random_circular_orbit, i=45),
                    wheelSpeeds=[1000.0, -1000.0, 1000.0],
                    nHat_B=np.array([0, 1, 0]),
                ),
            ),
            scenario=UniformTargets(n_targets=0),
            rewarder=data.NoReward(),
            sim_rate=1.0,
            max_step_duration=300.0,
            time_limit=1200.0,
            disable_env_checker=True,
        )

    def test_desat_action(self):
        env = self.make_env()
        env.reset()
        init_speeds = env.unwrapped.satellite.dynamics.wheel_speeds
        for _ in range(4):
            env.step(0)
            current_speeds = env.unwrapped.satellite.dynamics.wheel_speeds
            assert np.linalg.norm(current_speeds) < np.linalg.norm(init_speeds)
            init_speeds = current_speeds

    def test_desat_action_power_draw(self):
        env = self.make_env()
        env.unwrapped.satellite.sat_args_generator["thrusterPowerDraw"] = 0.0
        env.reset()
        env.step(0)  # Desat
        assert env.unwrapped.satellite.dynamics.battery_valid()

        env.unwrapped.satellite.sat_args_generator["thrusterPowerDraw"] = -10000.0
        env.reset()
        env.step(0)  # Desat
        assert not env.unwrapped.satellite.dynamics.battery_valid()

    def test_desat_action_pointing(self):
        env = self.make_env()
        env.unwrapped.satellite.sat_args_generator["desatAttitude"] = "nadir"
        env.reset(seed=0)
        env.step(0)  # Desat
        battery_level_nadir = env.unwrapped.satellite.dynamics.battery_charge

        env.unwrapped.satellite.sat_args_generator["desatAttitude"] = "sun"
        env.reset(seed=0)
        env.step(0)  # Desat
        battery_level_sun = env.unwrapped.satellite.dynamics.battery_charge
        assert battery_level_sun > battery_level_nadir


class TestNadirImagingActions:
    class ImageSat(sats.Satellite):
        dyn_type = dyn.ContinuousImagingDynModel
        fsw_type = fsw.ContinuousImagingFSWModel
        observation_spec = [obs.Time()]
        action_spec = [act.Scan()]

    env = gym.make(
        "SatelliteTasking-v1",
        satellite=ImageSat(
            "EO-1",
            sat_args=ImageSat.default_sat_args(
                oe=random_orbit,
                imageAttErrorRequirement=0.05,
                imageRateErrorRequirement=0.05,
                instrumentBaudRate=1.0,
                dataStorageCapacity=3.0,
                transmitterBaudRate=-1.0,
            ),
        ),
        scenario=UniformNadirScanning(),
        rewarder=data.NoReward(),
        sim_rate=1.0,
        time_limit=10000.0,
        max_step_duration=1e9,
        disable_env_checker=True,
    )

    def test_image(self):
        self.env.reset()
        storage_init = self.env.unwrapped.satellite.dynamics.storage_level
        self.env.step(0)
        assert self.env.unwrapped.satellite.dynamics.storage_level > storage_init



"""FSW models for RSO inspection."""

from typing import TYPE_CHECKING, Iterable

from Basilisk.architecture import messaging

from bsk_rl.sim.fsw import ContinuousImagingFSWModel, action
from bsk_rl.utils.functional import default_args

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.sats import Satellite


class RSOInspectorFSWModel(ContinuousImagingFSWModel):
    def set_target_rso(self, rso: "Satellite") -> None:
        """Set the RSO to point at."""
        self.locPoint.scTargetInMsg.subscribeTo(
            rso.dynamics.simpleNavObject.transOutMsg
        )

    class LocPointTask(ContinuousImagingFSWModel.LocPointTask):
        """Task to point at the RSO and trigger the instrument."""

        def __init__(self, *args, **kwargs) -> None:
            """Task to point at the RSO and trigger the instrument."""
            super().__init__(*args, **kwargs)

        @default_args(inst_pHat_B=[0, 0, 1])
        def setup_location_pointing(
            self, inst_pHat_B: Iterable[float], **kwargs
        ) -> None:
            """Set the location pointing guidance module to point at the RSO.

            The function ``set_target_rso`` must be called externally to connect the RSO
            to the pointing module.

            Args:
                inst_pHat_B: Instrument pointing direction.
                kwargs: Passed to other setup functions.
            """
            self.locPoint.pHat_B = inst_pHat_B
            self.locPoint.scAttInMsg.subscribeTo(
                self.fsw.dynamics.simpleNavObject.attOutMsg
            )
            self.locPoint.scTransInMsg.subscribeTo(
                self.fsw.dynamics.simpleNavObject.transOutMsg
            )
            self.locPoint.useBoresightRateDamping = 1
            messaging.AttGuidMsg_C_addAuthor(
                self.locPoint.attGuidOutMsg, self.fsw.attGuidMsg
            )

            self._add_model_to_task(self.locPoint, priority=1198)

        @default_args(imageAttErrorRequirement=0.01, imageRateErrorRequirement=None)
        def setup_instrument_controller(
            self,
            imageAttErrorRequirement: float,
            imageRateErrorRequirement: float,
            **kwargs,
        ) -> None:
            """Set the instrument controller parameters for scanning.

            As long as these two conditions are met, scanning will occur continuously.

            Args:
                imageAttErrorRequirement: [MRP norm] Pointing attitude error tolerance
                    for imaging.
                imageRateErrorRequirement: [rad/s] Rate tolerance for imaging. Disable
                    with None.
                kwargs: Passed to other setup functions.
            """
            self.insControl.attErrTolerance = imageAttErrorRequirement
            if imageRateErrorRequirement is not None:
                self.insControl.useRateTolerance = 1
                self.insControl.rateErrTolerance = imageRateErrorRequirement
            self.insControl.attGuidInMsg.subscribeTo(self.fsw.attGuidMsg)
            # Only use this module to check for pointing requirements
            self.access_msg = messaging.AccessMsg()
            payload = messaging.AccessMsgPayload()
            payload.hasAccess = 1
            self.access_msg.write(payload)
            self.insControl.accessInMsg.subscribeTo(self.access_msg)

            self._add_model_to_task(self.insControl, priority=987)

    @action
    def action_inspect_rso(self) -> None:
        """Action to inspect the RSO."""
        self.dynamics.instrument.nodeStatusInMsg.subscribeTo(
            self.insControl.deviceCmdOutMsg
        )
        self.insControl.controllerStatus = 1
        self.dynamics.instrumentPowerSink.powerStatus = 1
        self.dynamics.instrument.nodeDataName = "inspect_rso"
        self.simulator.enableTask(self.LocPointTask.name + self.satellite.name)


__doc_title__ = "RSO Inspection"
__all__ = ["RSOInspectorFSWModel"]



"""Classes for composing observations for a satellite."""

import logging
from abc import ABC, abstractmethod
from copy import deepcopy
from functools import cached_property
from typing import TYPE_CHECKING, Any, Union

import numpy as np
from gymnasium import spaces

from bsk_rl.utils.functional import Resetable, vectorize_nested_dict
from bsk_rl.utils.orbital import rv2HN

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.sats import Satellite
    from bsk_rl.sim import Simulator


logger = logging.getLogger(__name__)


def nested_obs_to_space(obs_dict, dtype):
    """Convert a nested observation dictionary to a gym space.

    Args:
        obs_dict: Observation dictionary
        dtype: Data type for observation np vectors.

    Returns:
        gym.Space: Observation space

    :meta private:
    """
    if isinstance(obs_dict, dict):
        return spaces.Dict(
            {
                key: nested_obs_to_space(value, dtype=dtype)
                for key, value in obs_dict.items()
            }
        )
    elif isinstance(obs_dict, list):
        return spaces.Box(low=-1e16, high=1e16, shape=(len(obs_dict),), dtype=dtype)
    elif isinstance(obs_dict, (float, int)):
        return spaces.Box(low=-1e16, high=1e16, shape=(1,), dtype=dtype)
    elif isinstance(obs_dict, np.ndarray):
        return spaces.Box(low=-1e16, high=1e16, shape=obs_dict.shape, dtype=dtype)
    else:
        raise TypeError(f"Cannot convert {obs_dict} to gym space.")


class ObservationBuilder(Resetable):
    def __init__(
        self,
        satellite: "Satellite",
        obs_type: type = np.ndarray,
        dtype: np.dtype = np.float64,
    ) -> None:
        """Satellite subclass for composing observations.

        Args:
            satellite: Satellite to observe
            obs_type: Datatype of satellite's returned observation. Can be ``np.ndarray``
                (default), ``dict``, or ``list``.
            dtype: Data type for observation np vectors.
        """
        self.obs_type = obs_type
        self.dtype = dtype
        self.obs_dict_cache = None
        self.obs_cache_time = 0.0
        self.satellite = satellite
        self.simulator: "Simulator"
        self.observation_spec = deepcopy(self.satellite.observation_spec)
        name_counts = {}
        for obs in self.observation_spec:
            if obs.name in name_counts:
                name_counts[obs.name] += 1
                obs.name += f"_{name_counts[obs.name]}"
            else:
                name_counts[obs.name] = 1
            obs.link_satellite(self.satellite)

    def reset_overwrite_previous(self) -> None:
        """Perform any once-per-episode setup."""
        for obs in self.observation_spec:
            obs.reset_overwrite_previous()

    def reset_pre_sim_init(self) -> None:
        """Perform any once-per-episode setup."""
        for obs in self.observation_spec:
            obs.reset_pre_sim_init()

    def reset_during_sim_init(self) -> None:
        """Perform any once-per-episode setup."""
        for obs in self.observation_spec:
            obs.reset_during_sim_init()

    def reset_post_sim_init(self) -> None:
        """Perform any once-per-episode setup."""
        self.simulator = self.satellite.simulator  # already a proxy
        self.obs_dict_cache = None
        for obs in self.observation_spec:
            obs.link_simulator(self.simulator)  # already a proxy
            obs.reset_post_sim_init()

    def obs_dict(self) -> dict[str, Any]:
        """Human-readable observation format.

        Cached so only computed once per timestep.
        """
        if (
            self.obs_dict_cache is None
            or self.simulator.sim_time != self.obs_cache_time
        ):
            self.obs_dict_cache = {}
            for obs_getter in self.observation_spec:
                obs = obs_getter.get_obs()
                if isinstance(obs, np.ndarray):
                    obs = obs.astype(self.dtype)
                self.obs_dict_cache[obs_getter.name] = obs
            self.obs_cache_time = self.simulator.sim_time
        return deepcopy(self.obs_dict_cache)

    def obs_ndarray(self) -> np.ndarray:
        """Numpy vector observation format."""
        _, obs = vectorize_nested_dict(self.obs_dict())
        return obs

    def obs_array_keys(self) -> list[str]:
        """Get the keys of the obs_ndarray."""
        keys, _ = vectorize_nested_dict(self.obs_dict())
        return keys

    def obs_list(self) -> list:
        """List observation format."""
        return list(self.obs_ndarray())

    def get_obs(self) -> Union[dict, np.ndarray, list]:
        """Update the observation."""
        if self.obs_type is dict:
            return self.obs_dict()
        elif self.obs_type is np.ndarray:
            return self.obs_ndarray()
        elif self.obs_type is list:
            return self.obs_list()
        else:
            raise ValueError(f"Invalid observation type: {self.obs_type}")

    @cached_property
    def observation_space(self) -> spaces.Space:
        """Space of the observation."""
        obs = self.get_obs()
        return nested_obs_to_space(obs, dtype=self.dtype)

    @cached_property
    def observation_description(self) -> Any:
        """Human-interpretable description of observation space."""
        return self.obs_array_keys()


class Observation(ABC, Resetable):
    """Base observations class."""

    def __init__(self, name: str = "obs") -> None:
        """Construct an observation.

        Args:
            name: Name of the observation.
        """
        self.name = name
        self.satellite: "Satellite"
        self.simulator: "Simulator"

    def link_satellite(self, satellite: "Satellite") -> None:
        """Link the observation to a satellite.

        Args:
            satellite: Satellite to link to

        :meta private:
        """
        self.satellite = satellite  # already a proxy

    def link_simulator(self, simulator: "Simulator") -> None:
        """Link the observation to a simulator.

        Args:
            simulator: Simulator to link to

        :meta private:
        """
        self.simulator = simulator  # already a proxy

    @abstractmethod  # pragma: no cover
    def get_obs(self) -> Any:
        """Return the observation."""
        pass


class SatProperties(Observation):
    """Add arbitrary `dynamics` and `fsw` ."""

    def __init__(self, *obs_properties: dict[str, Any], name="sat_props") -> None:
        """Include properties from ``fsw`` and ``dynamics`` in the observation.

        For each desired property, a dictionary specifying the property name and settings
        is passed. For example, to query the position and velocity of the satellite, the
        following would be used:

        .. code-block:: python

            SatProperties(
                dict(prop="r_BN_P", module="dynamics", norm=REQ_EARTH * 1e3),
                dict(prop="v_BN_P", module="dynamics", norm=7616.5, name="velocity"),
            ),


        Args:
            obs_properties: Property that can be found in fsw or dynamics that
                are to be appended to the the observation. Properties are optionally
                normalized by some factor. Each observation is a dictionary with the keys:

                * ``prop``: Name of property in ``fsw`` and ``dynamics`` to query
                * ``module`` `optional`: Module (dynamics or fsw) that holds the property. Can be inferred if ``None``.
                * ``norm`` `optional`: Value to normalize property by. Defaults to 1.0.
                * ``name`` `optional`: Name of the observation element. Defaults to the value of ``prop``.
                * ``fn`` `optional`: Alternatively, call a function that takes the satellite as an argument.
            name: Name of the observation.

        """
        super().__init__(name=name)
        for obs_property in obs_properties:
            for key in obs_property:
                if key not in ["prop", "module", "norm", "name", "fn"]:
                    raise ValueError(f"Invalid property key: {key}")
            if "norm" not in obs_property:
                obs_property["norm"] = 1.0
            if "name" not in obs_property:
                obs_property["name"] = obs_property["prop"]
                if obs_property["norm"] != 1.0:
                    obs_property["name"] += "_normd"

        self.obs_properties = obs_properties

    def reset_post_sim_init(self) -> None:
        """If necessary, automatically determine property location.

        :meta private:
        """
        for obs_property in self.obs_properties:
            if "module" not in obs_property and "fn" not in obs_property:
                for module in ["dynamics", "fsw"]:
                    if hasattr(getattr(self.satellite, module), obs_property["prop"]):
                        obs_property["module"] = module
                        break
                else:
                    raise AttributeError(f"Property {obs_property['prop']} not found")

    def get_obs(self) -> dict[str, Any]:
        """Return the observation.

        :meta private:
        """
        obs = {}
        for obs_property in self.obs_properties:
            prop = obs_property["prop"]
            if "fn" in obs_property:
                value = obs_property["fn"](self.satellite)
            else:
                module = obs_property["module"]
                value = getattr(getattr(self.satellite, module), prop)
            if isinstance(value, list):
                value = np.array(value)
            norm = obs_property["norm"]
            obs[obs_property["name"]] = value / norm
        return obs


class Time(Observation):
    def __init__(self, norm=None, name="time", observe_time_remaining=False):
        """Include the simulation time in the observation.

        Args:
            norm: Time to normalize by. If ``None``, the time is normalized by the simulation time limit.
            name: Name of the observation.
            observe_time_remaining: If ``True``, return the time remaining until the simulation time limit.
        """
        super().__init__(name=name)
        self.norm = norm
        self.return_time_remaining = observe_time_remaining

    def reset_post_sim_init(self) -> None:
        """Autodetect normalization time.

        :meta private:
        """
        if self.norm is None:
            self._norm = self.simulator.time_limit
        else:
            self._norm = self.norm

    def get_obs(self) -> float:
        """Return time normalized by normalization_time.

        :meta private:
        """
        if self.return_time_remaining:
            return (self.simulator.time_limit - self.simulator.sim_time) / self._norm
        else:
            return self.simulator.sim_time / self._norm


def _target_angle(sat, opp):
    vector_target_spacecraft_P = opp["r_LP_P"] - sat.dynamics.r_BN_P
    vector_target_spacecraft_P_hat = vector_target_spacecraft_P / np.linalg.norm(
        vector_target_spacecraft_P
    )
    return np.arccos(np.dot(vector_target_spacecraft_P_hat, sat.fsw.c_hat_P))


def _target_angle_rate(sat, opp):
    r_BN_P = sat.dynamics.v_BN_P
    v_BN_P = sat.dynamics.v_BN_P
    r_LP_P = opp["object"].r_LP_P
    omega_BP_P = sat.dynamics.omega_BP_P
    omega_CP_ref = (
        omega_BP_P
        - np.cross(v_BN_P, r_LP_P - r_BN_P) / np.linalg.norm(r_LP_P - r_BN_P) ** 2
    )
    return np.linalg.norm(omega_CP_ref)


def _r_LB_H(sat, opp):
    r_LP_P = opp["object"].r_LP_P
    r_BN_N = sat.dynamics.r_BN_N
    r_TB_N = sat.simulator.world.PN.T @ r_LP_P - r_BN_N
    HN = rv2HN(sat.dynamics.r_BN_N, sat.dynamics.v_BN_N)
    return HN @ r_TB_N


class OpportunityProperties(Observation):
    _fn_map = {
        "priority": lambda sat, opp: opp["object"].priority,
        "r_LP_P": lambda sat, opp: opp["r_LP_P"],
        "r_LB_H": _r_LB_H,
        "opportunity_open": lambda sat, opp: opp["window"][0] - sat.simulator.sim_time,
        "opportunity_mid": lambda sat, opp: sum(opp["window"]) / 2
        - sat.simulator.sim_time,
        "opportunity_close": lambda sat, opp: opp["window"][1] - sat.simulator.sim_time,
        "target_angle": _target_angle,
        "target_angle_rate": _target_angle_rate,
    }

    def __init__(
        self,
        *target_properties: dict[str, Any],
        n_ahead_observe: int,
        type="target",
        name=None,
    ):
        """Include information about upcoming access opportunities in the observation..

        For each desired property, a dictionary specifying the property name and settings
        is passed. These can include preset properties or arbitrary functions of the satellite
        and opportunity.

        .. code-block:: python

            OpportunityProperties(
                dict(prop="r_LP_P", norm=REQ_EARTH * 1e3),
                dict(prop="double_priority", fn=lambda sat, opp: opp["target"].priority * 2.0),
                n_ahead_observe=16,
            )

        Args:
            target_properties: Property that is a function of the opportunity to be appended
                to the the observation. Properties are optionally normalized by some factor.
                Each observation is a dictionary with the keys:

                * ``name`` `optional`: Name of the observation element.
                * ``fn`` `optional`: Function to calculate property, in the form ``fn(satellite, opportunity)``.
                  If not provided, the key ``prop`` will be used to look up a preset function:

                    * ``priority``: Priority of the target.
                    * ``r_LP_P``: Location of the target in the planet-fixed frame.
                    * ``r_LB_H``: Location of the target in the Hill frame.
                    * ``opportunity_open``: Time until the opportunity opens.
                    * ``opportunity_mid``: Time until the opportunity midpoint.
                    * ``opportunity_close``: Time until the opportunity closes.
                    * ``target_angle``: Angle between the target and the satellite instrument direction.
                    * ``target_angle_rate``: Rate difference between the target pointing frame and the body frame.

                * ``norm`` `optional`: Value to normalize property by. Defaults to 1.0.

            n_ahead_observe: Number of upcoming targets to consider.
            type: The type of opportunity to consider. Can be ``target``, ``ground_station``,
                or any other type of opportunity that has been added via
                :obj:`~bsk_rl.sats.AccessSatellite.add_location_for_access_checking`.
            name: Name of the observation.
        """
        if name is None:
            name = type
        super().__init__(name=name)
        self.type = type
        self.target_properties = target_properties
        for i, prop_spec in enumerate(self.target_properties):
            for key in prop_spec:
                if key not in ["fn", "norm", "name", "prop"]:
                    raise ValueError(f"Invalid property key: {key}")

            if "norm" not in prop_spec:
                prop_spec["norm"] = 1.0

            # Determine observation function
            if "fn" not in prop_spec:
                try:
                    prop_spec["fn"] = self._fn_map[prop_spec["prop"]]
                except KeyError:
                    raise ValueError(
                        f"Property prop={prop_spec['prop']} is not predefined and no `fn` was provided."
                    )
            else:
                if "prop" in prop_spec and prop_spec["prop"] in self._fn_map:
                    logger.warning(
                        f"Ignoring default function for `{prop_spec['prop']}` when `fn` is provided."
                    )

            # Determine best name
            if "name" not in prop_spec:
                if "prop" in prop_spec:
                    prop_spec["name"] = prop_spec["prop"]
                else:
                    prop_spec["name"] = f"prop_{i}"

                if prop_spec["norm"] != 1.0:
                    prop_spec["name"] += "_normd"

        self.n_ahead_observe = int(n_ahead_observe)

    def get_obs(self):
        """Iterate over property specs.

        :meta private:
        """
        from bsk_rl.sats import AccessSatellite

        if not isinstance(self.satellite, AccessSatellite):
            logger.warning(
                "OpportunityProperties observation requires an AccessSatellite"
            )

        obs = {}
        for i, opportunity in enumerate(
            self.satellite.find_next_opportunities(
                n=self.n_ahead_observe,
                types=self.type,
                pad=True,
            )
        ):
            props = {}
            for prop_spec in self.target_properties:
                name = prop_spec["name"]
                norm = prop_spec["norm"]
                value = prop_spec["fn"](self.satellite, opportunity)
                props[name] = value / norm
            obs[f"{self.name}_{i}"] = props
        return obs


class Eclipse(Observation):
    def __init__(self, norm=1.0, name="eclipse"):
        """Include a tuple of the next eclipse start and end times in the observation.

        Args:
            norm: Value to normalize by.
            name: Name of the observation.
        """
        super().__init__(name=name)
        self.norm = norm

    def get_obs(self):
        """Return tuple of normalized next eclipse start and end.

        :meta private:
        """
        eclipse_start, eclipse_end = self.satellite.trajectory.next_eclipse(
            self.simulator.sim_time
        )
        return [
            (eclipse_start - self.simulator.sim_time) / self.norm,
            (eclipse_end - self.simulator.sim_time) / self.norm,
        ]


class ResourceRewardWeight(Observation):
    """Observation for the weight of any :class:`ResourceReward`."""

    def __init__(
        self, name="resource_reward_weight", norm: Union[float, np.ndarray] = 1.0
    ):
        """Include the resource reward weight in the observation.

        Args:
            name: Name of the observation.
            norm: Value to normalize the resource reward weight by. If a vector, it should
                be the same length as the number of resource rewards in the environment.
        """
        super().__init__(name=name)
        self.norm = norm

    def reset_overwrite_previous(self) -> None:  # pragma: no cover
        """Prepare weights to be saved by rewarders."""
        self.weight_vector = []

    def get_obs(self) -> float:
        """Return the resource reward weight."""
        return np.array(self.weight_vector) / self.norm


__doc_title__ = "Backend"
__all__ = ["ObservationBuilder"]



"""``bsk_rl.sim`` is a package for the underlying Basilisk simulation.

The simulation is divided into three categories of Basilisk models:

* :ref:`bsk_rl.sim.world`, capturing elements of the simulation environment common to
  all satellites. This includes things such as gravity and atmosphere models, the epoch,
  and ground station locations. While the world model can be specified in the :class:`~bsk_rl.GeneralSatelliteTasking`
  constructor, it is generally automatically inferred from the satellite requirements.
* :ref:`bsk_rl.sim.dyn`, representing the dynamics model for each satellite. This is
  specified on a per-satellite basis by the :class:`~bsk_rl.sats.Satellite` type definition.
  The dynamics model captures the properties of the satellite, such as physical configurations,
  actuators models, instrument models, the power system, and storage resources.
* :ref:`bsk_rl.sim.fsw`, representing the flight software models for each satellite. As
  with flight software, this specified by the :class:`~bsk_rl.sats.Satellite`. The flight
  software model represents the low-level algorithms used for actuator and instrument
  control.

Generally, this can be thought of as a hierarchy of models, with dynamics models acting
in the world model, and flight software models controlling the dynamics models, and
other parts of ``bsk_rl`` controlling the flight software models. This hierarchy
contributes to the realism of the simulation, as the satellite is being controlled
through its flight software.

The :class:`~bsk_rl.sim.Simulator` is the main class for the simulation environment,
subclassing from the `Basilisk SimBaseClass <http://hanspeterschaub.info/basilisk/Documentation/utilities/SimulationBaseClass.html?highlight=simbaseclass#SimulationBaseClass.SimBaseClass>`_.
At each environment reset, the simulator and the associated flight software, dynamics,
and world models are deleted and reconstructed, generating a fresh Basilisk simulation.
"""

from bsk_rl.sim.simulator import Simulator

__doc_title__ = "Simulation (BSK)"
__all__ = [
    "Simulator",
]



from unittest.mock import MagicMock

import pytest

from bsk_rl.data.composition import ComposedData, ComposedDataStore, ComposedReward


class TestComposedData:
    @pytest.mark.parametrize(
        "data1, data2, expected",
        [
            ((1, 2), (3, 4), (4, 6)),
            ((), (3, 4), (3, 4)),
            ((1, 2), (), (1, 2)),
            ((), (), ()),
        ],
    )
    def test_add(self, data1, data2, expected):
        composed_data = ComposedData(*data1)
        other_composed_data = ComposedData(*data2)
        result = composed_data + other_composed_data
        assert result.data == expected

    def test_add_different_lengths(self):
        composed_data = ComposedData(1, 2)
        other_composed_data = ComposedData(3)
        with pytest.raises(ValueError):
            composed_data + other_composed_data

    def test_getattr(self):
        data1 = MagicMock()
        data1.a = 1
        del data1.b
        del data1.c
        data2 = MagicMock()
        data2.b = 2
        del data2.a
        del data2.c
        composed_data = ComposedData(data1, data2)
        assert composed_data.a == 1
        assert composed_data.b == 2
        with pytest.raises(AttributeError):
            _ = composed_data.c


class TestComposedDataStore:
    def test_pass_data(self):
        sat = MagicMock()
        ds1 = MagicMock()
        ds1_type = MagicMock(return_value=ds1)
        ds2 = MagicMock()
        ds2_type = MagicMock(return_value=ds2)
        composed_data_store = ComposedDataStore(sat, ds1_type, ds2_type)
        composed_data_store.data = ComposedData(1, 2)
        composed_data_store.pass_data()
        assert ds1.data == 1
        assert ds2.data == 2

    def test_getattr(self):
        sat = MagicMock()
        ds1 = MagicMock()
        ds1_type = MagicMock(return_value=ds1)
        ds2 = MagicMock()
        ds2_type = MagicMock(return_value=ds2)
        composed_data_store = ComposedDataStore(sat, ds1_type, ds2_type)
        ds1.a = 1
        assert composed_data_store.a == 1

    def test_get_log_state(self):
        sat = MagicMock()
        ds1 = MagicMock(get_log_state=MagicMock(return_value=1))
        ds1_type = MagicMock(return_value=ds1)
        ds2 = MagicMock(get_log_state=MagicMock(return_value=2))
        ds2_type = MagicMock(return_value=ds2)
        composed_data_store = ComposedDataStore(sat, ds1_type, ds2_type)

        log_states = composed_data_store.get_log_state()
        for ds in [ds1, ds2]:
            ds.get_log_state.assert_called_once()
        assert log_states == [1, 2]

    def test_compare_log_states(self):
        sat = MagicMock()
        ds1 = MagicMock(get_log_state=MagicMock(return_value=1))
        ds1_type = MagicMock(return_value=ds1)
        ds2 = MagicMock(get_log_state=MagicMock(return_value=2))
        ds2_type = MagicMock(return_value=ds2)
        composed_data_store = ComposedDataStore(sat, ds1_type, ds2_type)

        composed_data_store.compare_log_states([1, 2], [3, 4])
        ds1.compare_log_states.assert_called_once_with(1, 3)
        ds2.compare_log_states.assert_called_once_with(2, 4)


class TestComposedReward:
    def test_pass_data(self):
        rewarder1 = MagicMock()
        rewarder2 = MagicMock()
        composed_rewarder = ComposedReward(rewarder1, rewarder2)
        composed_rewarder.data = ComposedData(1, 2)
        composed_rewarder.pass_data()
        assert rewarder1.data == 1
        assert rewarder2.data == 2

    @pytest.mark.parametrize(
        "function",
        [
            "reset_pre_sim_init",
            "reset_post_sim_init",
            "reset_overwrite_previous",
        ],
    )
    def test_resetable(self, function):
        rewarder1 = MagicMock()
        rewarder2 = MagicMock()
        composed_rewarder = ComposedReward(rewarder1, rewarder2)
        getattr(composed_rewarder, function)()
        for rewarder in [rewarder1, rewarder2]:
            getattr(rewarder, function).assert_called_once()

    def test_initial_data(self):
        rewarder1 = MagicMock(
            initial_data=MagicMock(return_value=1),
        )
        rewarder2 = MagicMock(initial_data=MagicMock(return_value=2))
        composed_rewarder = ComposedReward(rewarder1, rewarder2)
        data = composed_rewarder.initial_data("sat")
        assert data.data == (1, 2)

    def test_data_store(self):
        sat = MagicMock()
        ds1 = MagicMock(get_log_state=MagicMock(return_value=1))
        ds1_type = MagicMock(return_value=ds1)
        rewarder1 = MagicMock(
            data_store_type=ds1_type,
            data_store_kwargs=dict(hello="world"),
        )
        ds2 = MagicMock(get_log_state=MagicMock(return_value=2))
        ds2_type = MagicMock(return_value=ds2)
        rewarder2 = MagicMock(data_store_type=ds2_type)
        composed_rewarder = ComposedReward(rewarder1, rewarder2)
        composed_rewarder.create_data_store(sat)
        assert sat.data_store.data_stores == (ds1, ds2)
        assert ds1_type.call_args.kwargs["hello"] == "world"
        assert "hello" not in ds2_type.call_args.kwargs

    def test_calculate_reward(self):
        rewarder1 = MagicMock(
            calculate_reward=MagicMock(return_value={"sat1": 1, "sat2": 2})
        )
        rewarder2 = MagicMock(
            calculate_reward=MagicMock(return_value={"sat1": 3, "sat2": 4})
        )
        composed_rewarder = ComposedReward(rewarder1, rewarder2)

        reward = composed_rewarder.calculate_reward(
            {
                "sat1": MagicMock(data=["d11", "d21"]),
                "sat2": MagicMock(data=["d12", "d22"]),
            }
        )

        assert reward == {"sat1": 4, "sat2": 6}



"""``bsk_rl.sats`` defines satellite agents for the environments.

A reference for working with satellites is given below. For a step-by-step guide, see
`this example <../../examples/satellite_configuration.ipynb>`_.

Configuring a Satellite
-----------------------

A subclass of a :class:`Satellite` type must be defined before can be used as an agent.
Two fields (``observation_spec`` and ``action_spec``) must be specified: these define
the observation and action spaces for the satellite. :ref:`bsk_rl.act` and :ref:`bsk_rl.obs`
provide more information on specifying the observation and action spaces.

Two other fields (``dyn_model`` and ``fsw_model``) may be specified to select the
underlying dynamics and FSW models used by the Basilisk simulation. Some actions,
communication methods, or other environment configurations may necessitate the use of a
specific dynamics or FSW model. See :ref:`bsk_rl.sim.fsw` and :ref:`bsk_rl.sim.dyn` for
more information on selecting these models.

In practice, configuring a satellite and passing it to an environment is straightforward:

.. code-block:: python

    class MySatellite(Satellite):
        observation_spec = [obs.Time(), ...]  # list of observations
        action_spec = [act.Drift(), ...]  # list of actions
        dyn_model = MyDynamicsModel  # dynamics model type
        fsw_model = MyFSWModel  # FSW model type

    my_sat = MySatellite(name="my_satellite")
    env = gym.make("SatelliteTasking-v1", satellite=my_sat, ...)


Setting Satellite Parameters
----------------------------

To specify satellite parameters such as physical properties and controller gains, a
``sat_args`` dictionary can be passed to the satellite constructor, which in turn is
used when initializing the FSW and dynamics simulations. Call the class method
:class:`~Satellite.default_sat_args` to list what parameters are available:

.. code-block:: python

    >>> MySatellite.default_sat_args()

    {'mass': 100.0, 'Kp': 0.1, 'Ki': 0.01, 'Kd': 0.01, ...}


These parameters are documented in :ref:`bsk_rl.sim.fsw` and :ref:`bsk_rl.sim.dyn`. To
override the default parameters, pass a dictionary with the desired values. Parameters
can be set by value, or by passing a function that returns a value. In the latter case,
the randomizer function will be called each time the simulation is reset. For example:

.. code-block:: python

    >>> my_sat = MySatellite(
            name="my_satellite",
            sat_args={"mass": lambda: np.random.uniform(95.0, 105.0), "Kp": 0.3}
        )
    >>> env = gym.make("SatelliteTasking-v1", satellite=my_sat, ...)
    >>> env.reset()
    >>> my_sat.sat_args

    {'mass': 98.372, 'Kp': 0.3, 'Ki': 0.01, 'Kd': 0.01, ...}

If one attempts to set a parameter that is not recognized, an error will be raised.

Helpful Methods for Debugging
-----------------------------
A variety of methods are available for debugging and introspection:

* :class:`~Satellite.observation_description` - Returns a human-interpretable description
  of the observation. For array-type observations, this can be useful to map indices to
  specific observation elements.
* :class:`~Satellite.action_description` - Returns a human-interpretable description of
  the actions. For discrete actions, this will be a list of action names.
* :class:`~Satellite.observation_space` - Returns the observation space for the single agent.
* :class:`~Satellite.action_space` - Returns the action space for the single agent.
* :class:`~Satellite.is_alive` - Returns whether the satellite is still operational based on
  ``@aliveness_checker`` s in the FSW and dynamics simulators.

Helpful Methods for Extending
-----------------------------
When extending the satellite class, certain convenience methods are available:

* :class:`~Satellite.reset_pre_sim_init` - Called on reset before the simulation is constructed.
* :class:`~Satellite.reset_post_sim_init` - Called on reset  after the simulation is constructed.
* ``logger.info/warning/debug`` - Logs a message to ``INFO``, associating it with the satellite.

Satellite Varieties
-------------------
* :class:`AccessSatellite` - Provides methods for determining when a satellite has
  access to a ground location based on elevation angle. Can return ordered lists of
  upcoming opportunities.
* :class:`ImagingSatellite` - Extends :class:`AccessSatellite` to provide methods for
  interacting with :class:`bsk_rl.scene.targets.Target` objects.
"""

from bsk_rl.sats.access_satellite import AccessSatellite, ImagingSatellite
from bsk_rl.sats.satellite import Satellite

__doc_title__ = "Satellites"
__all__ = ["Satellite", "AccessSatellite", "ImagingSatellite"]



"""Dynamics models concerning the relative motion of spacecraft."""

import types

import numpy as np
from Basilisk.simulation import spacecraftLocation
from Basilisk.utilities import macros

from bsk_rl.sim.dyn import BasicDynamicsModel
from bsk_rl.utils.functional import aliveness_checker, default_args, valid_func_name


class LOSCommDynModel(BasicDynamicsModel):
    """For evaluating line-of-sight connections between satellites for communication."""

    def __init__(self, *args, **kwargs) -> None:
        """Allow for line-of-sight checking between satellites.

        Necessary for :class:`~bsk_rl.comm.LOSCommunication` to function.
        """
        super().__init__(*args, **kwargs)

    def _setup_dynamics_objects(self, **kwargs) -> None:
        super()._setup_dynamics_objects(**kwargs)
        self.setup_los_comms(**kwargs)

    @default_args(losMaximumRange=-1.0)
    def setup_los_comms(
        self, losMaximumRange: float, priority: int = 500, **kwargs
    ) -> None:
        """Set up line-of-sight visibility checking between satellites.

        Args:
            losMaximumRange: [m] Maximum range for line-of-sight visibility. -1 for unlimited.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.losComms = spacecraftLocation.SpacecraftLocation()
        self.losComms.ModelTag = "losComms"
        self.losComms.primaryScStateInMsg.subscribeTo(self.scObject.scStateOutMsg)
        self.losComms.planetInMsg.subscribeTo(
            self.world.gravFactory.spiceObject.planetStateOutMsgs[self.world.body_index]
        )
        self.losComms.rEquator = self.simulator.world.planet.radEquator
        self.losComms.rPolar = self.simulator.world.planet.radEquator * 0.98
        self.losComms.maximumRange = losMaximumRange

        self.los_comms_ids = []

        for sat_dyn in self.simulator.dynamics_list.values():
            if sat_dyn != self and sat_dyn.satellite.name not in self.los_comms_ids:
                self.losComms.addSpacecraftToModel(sat_dyn.scObject.scStateOutMsg)
                self.los_comms_ids.append(sat_dyn.satellite.name)
                sat_dyn.losComms.addSpacecraftToModel(self.scObject.scStateOutMsg)
                sat_dyn.los_comms_ids.append(self.satellite.name)
                if len(sat_dyn.los_comms_ids) == 1:
                    sat_dyn.simulator.AddModelToTask(
                        sat_dyn.task_name, sat_dyn.losComms, ModelPriority=priority
                    )

        if len(self.los_comms_ids) > 0:
            self.simulator.AddModelToTask(
                self.task_name, self.losComms, ModelPriority=priority
            )


class ConjunctionDynModel(BasicDynamicsModel):
    """For evaluating conjunctions between satellites."""

    def __init__(self, *args, **kwargs) -> None:
        """Model that evaluates conjunctions between satellites.

        The simulation is terminated at the time of collision and a conjunction_valid failure is reported.
        """
        super().__init__(*args, **kwargs)
        self.conjunctions = []

    def _setup_dynamics_objects(self, **kwargs) -> None:
        super()._setup_dynamics_objects(**kwargs)
        self.setup_conjunctions(**kwargs)

    @aliveness_checker
    def conjunction_valid(self) -> bool:
        """Check if conjunction has not occured."""
        return len(self.conjunctions) == 0

    @default_args(conjunction_radius=10)
    def setup_conjunctions(self, conjunction_radius: float, **kwargs) -> None:
        """Set up conjunction checking between satellites.

        Args:
            conjunction_radius: [m] Minimum distance for a conjunction.
            kwargs: Passed to other setup functions.
        """
        self.conjunction_radius = conjunction_radius

        for sat_dyn in self.simulator.dynamics_list.values():
            if sat_dyn != self and isinstance(sat_dyn, ConjunctionDynModel):

                def condition(sim, sat_dyn=sat_dyn):
                    distance = np.linalg.norm(
                        np.array(self.satellite.dynamics.r_BN_N)
                        - np.array(sat_dyn.satellite.dynamics.r_BN_N)
                    )
                    keepout = (
                        self.satellite.dynamics.conjunction_radius
                        + sat_dyn.satellite.dynamics.conjunction_radius
                    )
                    return distance <= keepout

                def side_effect(sim):
                    self.satellite.logger.info(
                        f"collided with {sat_dyn.satellite.name}"
                    )
                    sat_dyn.satellite.logger.info(
                        f"collided with {self.satellite.name}"
                    )
                    self.satellite.dynamics.conjunctions.append(sat_dyn.satellite)
                    sat_dyn.satellite.dynamics.conjunctions.append(self.satellite)
                    if sim.sim_time == 0:
                        self.satellite.logger.warning(
                            "Collision occurred at t=0, may incorrectly report failure type"
                        )

                self.simulator.createNewEvent(
                    valid_func_name(
                        f"conjunction_{self.satellite.name}_{sat_dyn.satellite.name}"
                    ),
                    macros.sec2nano(self.simulator.sim_rate),
                    True,
                    conditionFunction=condition,
                    actionFunction=side_effect,
                    terminal=True,
                )


class MaxRangeDynModel(BasicDynamicsModel):
    """For evaluating a maximum range limitation between satellites."""

    def __init__(self, *args, **kwargs) -> None:
        """Model that checks for maximum range violations between satellites.

        The simulation is terminated at the time of separation and a range_valid failure is reported.
        """
        super().__init__(*args, **kwargs)
        self.out_of_ranges = []

    def _setup_dynamics_objects(self, **kwargs) -> None:
        super()._setup_dynamics_objects(**kwargs)
        self.setup_range(**kwargs)

    @aliveness_checker
    def range_valid(self) -> bool:
        """Check if conjunction has not occurred."""
        return len(self.out_of_ranges) == 0

    @default_args(max_range_radius=5000, chief_name=None, enforce_range_on_chief=False)
    def setup_range(
        self,
        max_range_radius: float,
        chief_name: str,
        enforce_range_on_chief: bool,
        **kwargs,
    ) -> None:
        """Set up maximum distance checking relative to a chief satellite.

        Args:
            max_range_radius: [m] Maximum allowed range from the chief satellite.
            chief_name: Chief satellite to check range against.
            enforce_range_on_chief: If True, the chief will also die if the range is
                violated by this satellite.
            kwargs: Passed to other setup functions.
        """
        self.max_range_radius = max_range_radius
        self.chief_name = chief_name

        if self.chief_name is None:
            self.logger.warning(
                "No chief satellite specified for maximum range checking. "
                "Range checking is disabled."
            )
            return

        self.chief = self.simulator.get_satellite(self.chief_name)

        # Add range check to chief if required
        if enforce_range_on_chief:

            @aliveness_checker
            def range_valid(self, deputy=self):
                return deputy.range_valid()

            setattr(
                self.chief.dynamics,
                valid_func_name(f"range_valid_{self.satellite.name}"),
                types.MethodType(range_valid, self.chief.dynamics),
            )

        # Add event to check for max range violation
        def condition(sim):
            distance = np.linalg.norm(
                np.array(self.satellite.dynamics.r_BN_N)
                - np.array(self.chief.dynamics.r_BN_N)
            )
            return distance >= self.max_range_radius

        def side_effect(sim):
            self.satellite.logger.info(
                f"Exceeded maximum range of {max_range_radius} m from {self.chief_name}"
            )
            self.out_of_ranges.append(self.chief)

        self.simulator.createNewEvent(
            valid_func_name(f"range_{self.satellite.name}_{self.chief_name}"),
            macros.sec2nano(self.simulator.sim_rate),
            True,
            conditionFunction=condition,
            actionFunction=side_effect,
            terminal=True,
        )


__doc_title__ = "Relative Motion"
__all__ = [
    "LOSCommDynModel",
    "ConjunctionDynModel",
    "MaxRangeDynModel",
]
